/* Copyright (c) 2013-2015 Richard Rodger, MIT License */

/* jshint node:true, asi:true, eqnull:true */
"use strict";

var _ = require('lodash');

var common = require('./common');

exports.log_act_in = function (instance, actinfo, actmeta, args, prior_ctxt) {
  if (actmeta.sub || !instance.log.levels.debug) return;
  instance.log.debug('act', actmeta.plugin_name, actmeta.plugin_tag, 'IN', actinfo.actid, actmeta.pattern, function () {
    return [actmeta.descdata ? actmeta.descdata(args) : common.descdata(args), prior_ctxt.entry ? 'ENTRY' : 'PRIOR;' + prior_ctxt.chain.slice(0, prior_ctxt.depth).join(','), 'A;' + actmeta.id, args.gate$ ? 'GATE' : '-'];
  });
};

exports.log_act_out = function (instance, actinfo, actmeta, args, result, prior_ctxt) {
  if (actmeta.sub || !instance.log.levels.debug) return;
  instance.log.debug('act', actmeta.plugin_name, actmeta.plugin_tag, 'OUT', actinfo.actid, actmeta.pattern, function () {
    return _.flatten([_.flatten([actmeta.descdata ? actmeta.descdata(result.slice(1)) : common.descdata(result.slice(1))], true), prior_ctxt.entry ? 'EXIT' : 'PRIOR;' + prior_ctxt.chain.slice(0, prior_ctxt.depth).join(','), 'A;' + actmeta.id, actinfo.duration, args.gate$ ? 'GATE' : '-']);
  });
};

exports.log_act_err = function (instance, actinfo, actmeta, args, prior_ctxt, err) {
  if (!err.log || !instance.log.levels.error) return;
  instance.log.error('act', actmeta.plugin_name || '-', actmeta.plugin_tag || '-', 'OUT', actinfo.actid, actmeta.pattern || '-', actinfo.duration, actmeta.descdata ? actmeta.descdata(args) : common.descdata(args), prior_ctxt.entry ? 'ENTRY' : 'PRIOR;' + prior_ctxt.chain.slice(0, prior_ctxt.depth).join(','), 'A;' + actmeta.id, args.gate$ ? 'GATE' : '-', err.message, err.code, common.descdata(err.details), err.stack, args.caller$);
};

exports.log_act_cache = function (instance, actinfo, actmeta, args, prior_ctxt) {
  if (!instance.log.levels.debug) return;
  instance.log.debug('act', actmeta.plugin_name, actmeta.plugin_tag, 'OUT', actinfo.actid, actmeta.pattern, 'CACHE', prior_ctxt.entry ? 'ENTRY' : 'PRIOR;' + prior_ctxt.chain.slice(0, prior_ctxt.depth).join(','), function () {
    return [actmeta.descdata ? actmeta.descdata(args) : common.descdata(args), 'A=' + actmeta.id];
  });
};

exports.log_exec_err = function (instance, err) {
  if (!err.log || instance.log.levels.error) return;
  err.details = err.details || {};
  err.details.plugin = err.details.plugin || {};
  instance.log.error('act', err.details.plugin.name || '-', err.details.plugin.tag || '-', err.details.id || '-', err.details.pattern || '-', err.message, err.code, common.descdata(err.details), err.stack);
};

exports.log_act_not_found = function (instance, err, loglevel) {
  if (false === err.log) return;
  loglevel = loglevel || 'warn';
  if ('ignore' == loglevel) return;
  err.details = err.details || {};
  err.details.plugin = err.details.plugin || {};

  if (instance.log.levels[loglevel]) {
    instance.log(loglevel, 'act', err.details.plugin.name || '-', err.details.plugin.tag || '-', err.details.id || '-', err.details.pattern || '-', err.message, err.code, common.descdata(err.details), err.stack);
  }
};

var noop = function noop() {};

var levels = ['debug', // 0
'info', // 1
'warn', // 2
'error', // 3
'fatal' // 4
];

var normalizeLevel = function normalizeLevel(level) {
  level = level.toLowerCase();

  if (levels.indexOf(level) === -1 && level !== 'silent') {
    throw new Error('Invalid log level ' + level);
  }

  return level;
};

var parseLevel = function parseLevel(options) {
  if (options.log && options.log.level) {
    return normalizeLevel(options.log.level);
  }

  if (options.log && typeof options.log === 'string') {
    return normalizeLevel(options.log);
  }

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = process.argv.slice(2)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var arg = _step.value;
      var match = arg.match(/--seneca\.log=level:(.*)/);

      if (match) {
        return normalizeLevel(match[1]);
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return process.env.SENECA_LOG ? normalizeLevel(process.env.SENECA_LOG) : 'warn';
};

exports.createLogger = function (options) {
  var level = parseLevel(options);
  var handler = options.log && options.log.handler || console.log;

  var log = function log(level) {
    if (log.levels[level]) {
      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      handler.apply(void 0, [level].concat(args));
    }
  };

  log.levels = {};
  levels.forEach(function (currentLevel, index) {
    if (index + 1 > levels.indexOf(level) && level !== 'silent') {
      log[currentLevel] = function () {
        for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
          args[_key2] = arguments[_key2];
        }

        return handler.apply(void 0, [currentLevel].concat(args));
      };

      log.levels[currentLevel] = true;
    } else {
      log[currentLevel] = noop;
      log.levels[currentLevel] = false;
    }
  });
  Object.freeze(log);
  return log;
};